/*
 * EmulatorUILogic.java
 *
 * Created on April 16, 2007, 10:30 PM
 */
package jace;

import jace.apple2e.MOS65C02;
import jace.config.InvokableAction;
import jace.core.CPU;
import jace.core.Computer;
import jace.core.Debugger;
import jace.core.RAM;
import jace.core.RAMEvent;
import jace.core.RAMListener;
import static jace.core.Utility.*;
import jace.ui.AbstractEmulatorFrame;
import jace.ui.DebuggerPanel;
import java.awt.Color;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JTextField;

public class EmulatorUILogic {

    static Debugger debugger;

    static {
        debugger = new Debugger() {
            @Override
            public void updateStatus() {
                enableDebug(true);
                MOS65C02 cpu = (MOS65C02) Computer.getComputer().getCpu();
                updateCPURegisters(cpu);
            }
        };
    }

    public static void updateCPURegisters(MOS65C02 cpu) {
        DebuggerPanel debuggerPanel = Emulator.getFrame().getDebuggerPanel();
        debuggerPanel.valueA.setText(Integer.toHexString(cpu.A));
        debuggerPanel.valueX.setText(Integer.toHexString(cpu.X));
        debuggerPanel.valueY.setText(Integer.toHexString(cpu.Y));
        debuggerPanel.valuePC.setText(Integer.toHexString(cpu.getProgramCounter()));
        debuggerPanel.valueSP.setText(Integer.toHexString(cpu.getSTACK()));
        debuggerPanel.valuePC2.setText(cpu.getFlags());
        debuggerPanel.valueINST.setText(cpu.disassemble());
    }

    public static void enableDebug(boolean b) {
        DebuggerPanel debuggerPanel = Emulator.getFrame().getDebuggerPanel();
        debugger.setActive(b);
        debuggerPanel.enableDebug.setSelected(b);
        debuggerPanel.setBackground(
                b ? Color.RED : new Color(0, 0, 0x040));
    }

    public static void enableTrace(boolean b) {
        Computer.getComputer().getCpu().setTraceEnabled(b);
    }

    public static void stepForward() {
        debugger.step = true;
    }

    static void registerDebugger() {
        Computer.getComputer().getCpu().setDebug(debugger);
    }

    public static Integer getValidAddress(String s) {
        try {
            int addr = Integer.parseInt(s.toUpperCase(), 16);
            if (addr >= 0 && addr < 0x10000) {
                return addr;
            }
            return null;
        } catch (NumberFormatException ex) {
            return null;
        }
    }
    public static List<RAMListener> watches = new ArrayList<RAMListener>();

    public static void updateWatchList(final DebuggerPanel panel) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            @Override
            public void run() {
                // Remove old listeners
                for (RAMListener oldWatch : watches) {
                    Computer.getComputer().getMemory().removeListener(oldWatch);
                }
                if (panel == null) {
                    return;
                }
                addWatch(panel.textW1, panel.valueW1);
                addWatch(panel.textW2, panel.valueW2);
                addWatch(panel.textW3, panel.valueW3);
                addWatch(panel.textW4, panel.valueW4);
            }
        });
    }

    private static void addWatch(JTextField watch, final JLabel watchValue) {
        final Integer address = getValidAddress(watch.getText());
        if (address != null) {
            //System.out.println("Adding watch for "+Integer.toString(address, 16));
            RAMListener newListener = new RAMListener(RAMEvent.TYPE.WRITE, RAMEvent.SCOPE.ADDRESS, RAMEvent.VALUE.ANY) {
                @Override
                protected void doConfig() {
                    setScopeStart(address);
                }

                @Override
                protected void doEvent(RAMEvent e) {
                    watchValue.setText(Integer.toHexString(e.getNewValue() & 0x0FF));
                }
            };
            Computer.getComputer().getMemory().addListener(newListener);
            watches.add(newListener);
            // Print out the current value right away
            byte b = Computer.getComputer().getMemory().readRaw(address);
            watchValue.setText(Integer.toString(b & 0x0ff, 16));
        } else {
            watchValue.setText("00");
        }
    }


    public static void updateBreakpointList(final DebuggerPanel panel) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            @Override
            public void run() {
                Integer address;
                debugger.getBreakpoints().clear();
                if (panel == null) {
                    return;
                }
                address = getValidAddress(panel.textBP1.getText());
                if (address != null) {
                    debugger.getBreakpoints().add(address);
                }
                address = getValidAddress(panel.textBP2.getText());
                if (address != null) {
                    debugger.getBreakpoints().add(address);
                }
                address = getValidAddress(panel.textBP3.getText());
                if (address != null) {
                    debugger.getBreakpoints().add(address);
                }
                address = getValidAddress(panel.textBP4.getText());
                if (address != null) {
                    debugger.getBreakpoints().add(address);
                }
                debugger.updateBreakpoints();
            }
        });
    }

    @InvokableAction(
            name = "BRUN file",
    category = "file",
    description = "Loads a binary file in memory and executes it. File should end with #06xxxx, where xxxx is the start address in hex",
    alternatives = "Execute program;load binary;load program;load rom;play single-load game")
    public static void runFile() {
        Computer.getComputer().getCpu().suspend();
        JFileChooser select = new JFileChooser();
        select.showDialog(Emulator.getFrame(), "Execute binary file");
        File binary = select.getSelectedFile();
        if (binary == null) {
            Computer.getComputer().getCpu().resume();
            return;
        }
        runFile(binary);
    }

    public static void runFile(File binary) {
        String fileName = binary.getName().toLowerCase();
        try {
            if (fileName.contains("#06")) {
                String addressStr = fileName.substring(fileName.length() - 4);
                int address = Integer.parseInt(addressStr, 16);
                brun(binary, address);
            } else if (fileName.contains("#fc")) {
                gripe("BASIC not supported yet");
            }
        } catch (NumberFormatException ex) {
            ex.printStackTrace();
        } catch (FileNotFoundException ex) {
            ex.printStackTrace();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        Computer.getComputer().getCpu().resume();
    }

    public static void brun(File binary, int address) throws FileNotFoundException, IOException {
        // If it was halted already, then it was initiated outside of an opcode execution
        // If it was not yet halted, then it is the case that the CPU is processing another opcode
        // So if that is the case, the program counter will need to be decremented here to compensate
        // TODO: Find a better mousetrap for this one -- it's an ugly hack
        Computer.getComputer().getCpu().suspend();
        FileInputStream in = new FileInputStream(binary);
        byte[] data = new byte[in.available()];
        in.read(data);
        RAM ram = Computer.getComputer().getMemory();
        for (int i = 0; i < data.length; i++) {
            ram.write(address + i, data[i], false);
        }
        CPU cpu = Computer.getComputer().getCpu();
        Computer.getComputer().getCpu().setProgramCounter(address);
        Computer.getComputer().getCpu().resume();
    }

    @InvokableAction(
            name = "Adjust display",
    category = "display",
    description = "Adjusts window size to 1:1 aspect ratio for optimal viewing.",
    alternatives = "Adjust screen;Adjust window size;Adjust aspect ratio;Fix screen;Fix window size;Fix aspect ratio;Correct aspect ratio;")
    static public void scaleIntegerRatio() {
        AbstractEmulatorFrame frame = Emulator.getFrame();
        if (frame == null) return;
        Computer.pause();
        frame.enforceIntegerRatio();
        Computer.resume();
    }

    public static void toggleDebugPanel() {
        AbstractEmulatorFrame frame = Emulator.getFrame();
        if (frame == null) return;
        frame.setShowDebug(!frame.isShowDebug());
        frame.reconfigure();
        Emulator.resizeVideo();
    }

    public static void toggleFullscreen() {
        AbstractEmulatorFrame frame = Emulator.getFrame();
        if (frame == null) return;
        Computer.pause();
        frame.toggleFullscreen();
        Computer.resume();
    }
}
