package jace.core;

import jace.apple2e.SoftSwitches;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

/**
 *
 * @author blurry
 */
public abstract class RAM {

    protected PagedMemory activeRead;
    protected PagedMemory activeWrite;
    private List<RAMListener> listeners;
    private List<RAMListener>[] listenerMap;
    private List<RAMListener>[] ioListenerMap;
    Card[] cards;
    // card 0 = 80 column card firmware / system rom
    int activeSlot = 0;

    /**
     * Creates a new instance of RAM
     */
    public RAM() {
        listeners = new Vector<RAMListener>();
        cards = new Card[8];
        refreshListenerMap();
    }

    public void setActiveCard(int slot) {
        if (activeSlot != slot) {
            activeSlot = slot;
            configureActiveMemory();
        } else if (!SoftSwitches.CXROM.getState()) {
            configureActiveMemory();
        }
    }

    public int getActiveSlot() {
        return activeSlot;
    }

    public Card[] getAllCards() {
        return cards;
    }

    public Card getCard(int slot) {
        if (slot >= 1 && slot <= 7) {
            return cards[slot];
        }
        return null;
    }

    public void addCard(Card c, int slot) {
        cards[slot] = c;
        c.setSlot(slot);
        c.attach();
    }

    public void removeCard(Card c) {
        c.suspend();
        c.detach();
        removeCard(c.getSlot());
    }

    public void removeCard(int slot) {
        if (cards[slot] != null) {
            cards[slot].suspend();
            cards[slot].detach();
        }
        cards[slot] = null;
    }

    abstract public void configureActiveMemory();

    public byte write(int address, byte b, boolean generateEvent) {
        byte[] page = activeWrite.getMemoryPage(address);
        byte old = 0;
        if (page == null) {
            if (generateEvent) {
                callListener(RAMEvent.TYPE.WRITE, address, old, b);
            }
        } else {
            int offset = address & 0x0FF;
            old = page[offset];
            if (generateEvent) {
                b = callListener(RAMEvent.TYPE.WRITE, address, old, b);
            }
            page[offset] = b;
        }
        return old;
    }

    public void writeWord(int address, int w, boolean generateEvent) {
        int lsb = write(address, (byte) (w & 0x0ff), generateEvent);
        int msb = write(address + 1, (byte) (w >> 8), generateEvent);
//        int oldValue = msb << 8 + lsb;
    }

    public byte readRaw(int address) {
        //    if (address >= 65536) return 0;
        return activeRead.getMemoryPage(address)[address & 0x0FF];
    }

    public byte read(int address, RAMEvent.TYPE eventType, boolean triggerEvent) {
        //    if (address >= 65536) return 0;
        byte value = activeRead.getMemoryPage(address)[address & 0x0FF];
//        if (triggerEvent || ((address & 0x0FF00) == 0x0C000)) {
        if (triggerEvent || (address & 0x0FFF0) == 0x0c030) {
            value = callListener(eventType, address, value, value);
        }
        return value;
    }

    public int readWordRaw(int address) {
        int lsb = 0x00ff & readRaw(address);
        int msb = (0x00ff & readRaw(address + 1)) << 8;
        return msb + lsb;
    }

    public int readWord(int address, RAMEvent.TYPE eventType, boolean triggerEvent) {
        int lsb = 0x00ff & read(address, eventType, triggerEvent);
        int msb = (0x00ff & read(address + 1, eventType, triggerEvent)) << 8;
        int value = msb + lsb;
//        if (generateEvent) {
//            callListener(RAMEvent.TYPE.READ, address, value, value);
//        }
        return value;
    }

    private void mapListener(RAMListener l, int address) {
        if ((address & 0x0FF00) == 0x0C000) {
            int index = address & 0x0FF;
            List<RAMListener> listeners = ioListenerMap[index];
            if (listeners == null) {
                listeners = new ArrayList<RAMListener>();
                ioListenerMap[index] = listeners;
            }
            if (!listeners.contains(l)) {
                listeners.add(l);
            }
        } else {
            int index = address >> 8;
            List<RAMListener> listeners = listenerMap[index];
            if (listeners == null) {
                listeners = new ArrayList<RAMListener>();
                listenerMap[index] = listeners;
            }
            if (!listeners.contains(l)) {
                listeners.add(l);
            }
        }
    }

    private void addListenerRange(RAMListener l) {
        if (l.getScope() == RAMEvent.SCOPE.ADDRESS) {
            mapListener(l, l.getScopeStart());
        } else {
            int start = 0;
            int end = 0x0ffff;
            if (l.getScope() == RAMEvent.SCOPE.RANGE) {
                start = l.getScopeStart();
                end = l.getScopeEnd();
            }
            for (int i = start; i <= end; i++) {
                mapListener(l, i);
            }
        }
    }

    private void refreshListenerMap() {
        listenerMap = new ArrayList[256];
        ioListenerMap = new ArrayList[256];
        for (RAMListener l : listeners) {
            addListenerRange(l);
        }
    }

    public void addListener(final RAMListener l) {
        boolean restart = Computer.pause();
        if (listeners.contains(l)) {
            return;
        }
        listeners.add(l);
        addListenerRange(l);
        if (restart) {Computer.resume();}
    }

    public void removeListener(final RAMListener l) {
        boolean restart = Computer.pause();
        listeners.remove(l);
        refreshListenerMap();
        if (restart) {Computer.resume();}
    }

    public byte callListener(RAMEvent.TYPE t, int address, int oldValue, int newValue) {
        List<RAMListener> activeListeners = null;
        Computer.getComputer().getCpu().suspend();
        if ((address & 0x0FF00) == 0x0C000) {
            activeListeners = ioListenerMap[address & 0x0FF];
            if (activeListeners == null && t.isRead()) {
                Computer.getComputer().getCpu().resume();
                return Computer.getComputer().getVideo().getFloatingBus();
            }
        } else {
            activeListeners = listenerMap[(address >> 8) & 0x0ff];
        }
        if (activeListeners != null) {
            RAMEvent e = new RAMEvent(t, RAMEvent.SCOPE.ADDRESS, RAMEvent.VALUE.ANY, address, oldValue, newValue);
            for (RAMListener l : activeListeners) {
                l.handleEvent(e);
            }
            Computer.getComputer().getCpu().resume();
            return (byte) e.getNewValue();
        }
        Computer.getComputer().getCpu().resume();
        return (byte) newValue;
    }

    abstract protected void loadRom(String path) throws IOException;

    public void dump() {
        for (int i = 0; i < 0x0FFFF; i += 16) {
            System.out.print(Integer.toString(i, 16));
            System.out.print(":");
            String part1 = "";
            String part2 = "";
            for (int j = 0; j < 16; j++) {
                int a = i + j;
                int br = 0x0FF & activeRead.getMemory()[i >> 8][i & 0x0ff];
                String s1 = Integer.toString(br, 16);
                System.out.print(' ');
                if (s1.length() == 1) {
                    System.out.print('0');
                }
                System.out.print(s1);

                /*
                 try {
                 int bw = 0;
                 bw = 0x0FF & activeWrite.getMemory().get(a/256)[a%256];
                 String s2 = (br == bw) ? "**" : Integer.toString(bw,16);
                 System.out.print(' ');
                 if (s2.length()==1) System.out.print('0');
                 System.out.print(s2);
                 } catch (NullPointerException ex) {
                 System.out.print(" --");
                 }
                 */
            }
            System.out.println();
//            System.out.println(Integer.toString(i, 16)+":"+part1+" -> "+part2);
        }
    }
}