package jace.core;

import jace.config.ConfigurableField;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 *
 * @author blurry
 */
public abstract class TimedDevice extends Device {

    /**
     * Creates a new instance of TimedDevice
     */
    public TimedDevice() {
        setSpeed(cyclesPerSecond);
        isPaused = true;
    }
    @ConfigurableField(name = "Speed", description = "(in hertz)")
    public long cyclesPerSecond = defaultCyclesPerSecond();
    @ConfigurableField(name = "Max speed")
    public boolean maxspeed = false;

    @Override
    public abstract void tick();
    private static final double NANOS_PER_SECOND = 1000000000.0;
    // current cycle within the period
    private int cycleTimer = 0;
    // The actual worker that the device runs as
    public Thread worker;
    public static int TEMP_SPEED_MAX_DURATION = 10000000;
    private int tempSpeedDuration = 0;

    @Override
    public boolean suspend() {
        disableTempMaxSpeed();
        boolean result = super.suspend();
        if (worker == null || !worker.isAlive()) {
            return result;
        }
        if (worker == Thread.currentThread()) {
            return result;
        }
        worker.interrupt();
        try {
            pauseLock.lock();
            while (!isPaused) {
                pausedCondition.await();
            }
        } catch (Exception e) {
        } finally {
            pauseLock.unlock();
        }
        return result;
    }
    Lock pauseLock = new ReentrantLock();
    Condition resumeCondition = pauseLock.newCondition();
    Condition pausedCondition = pauseLock.newCondition();
    boolean haltRequested = false;
    Thread clock;
    Lock clockLock = new ReentrantLock();
    Condition clockPulse = clockLock.newCondition();

    public boolean isHalted() {
        return worker == null || !worker.isAlive();
    }

    public void halt() {
        haltRequested = true;
        clock.interrupt();
        suspend();
    }

    @Override
    public void resume() {
        haltRequested = false;
        super.resume();
        try {
            pauseLock.lock();
            resumeCondition.signalAll();
        } catch (Exception ex) {
        } finally {
            pauseLock.unlock();
        }
        if (worker != null && worker.isAlive()) {
            return;
        }

        try {
            clockLock.lock();
            if (clock == null || !clock.isAlive()) {
                clock = new Thread(new Runnable() {
                    @Override
                    public void run() {
                        long nextSync = System.nanoTime() + nanosPerInterval;
                        while (!haltRequested) {
                            long now = System.nanoTime();
                            if (now < nextSync) {
                                long currentSyncDiff = nextSync - now;
                                if (currentSyncDiff > 1000000L) {
                                    try {
                                        Thread.sleep(currentSyncDiff / 1000000L, (int) (currentSyncDiff % 1000000L));
                                    } catch (InterruptedException ex) {
                                        // Probably being suspended, no worries
                                    }
                                }
                            }
                            if (haltRequested) {
                                break;
                            }
                            nextSync += nanosPerInterval;
                            try {
                                clockLock.lock();
                                clockPulse.signalAll();
                            } finally {
                                clockLock.unlock();
                            }
                        }
                        try {
                            clockLock.lock();
                            clock = null;
                        } finally {
                            clockLock.unlock();
                        }
                    }
                }, "Timed device clock " + getName());
                clock.setPriority(Thread.MAX_PRIORITY);
                clock.start();
            }
        } finally {
            clockLock.unlock();
        }
        worker = new Thread(new Runnable() {
            @Override
            public void run() {
                while (!haltRequested) {
                    isPaused = false;
                    while (isRunning()) {
                        doTick();
                        resync();
                    }
                    isPaused = true;
                    if (haltRequested) {
                        break;
                    }
                    try {
                        pauseLock.lock();
                        pausedCondition.signalAll();
                        while (!isRunning()) {
                            resumeCondition.await(1000, TimeUnit.MILLISECONDS);
                        }
                    } catch (InterruptedException ex) {
                    } finally {
                        pauseLock.unlock();
                    }
                }
                worker = null;
                System.out.println("Halting " + getName());
            }
        }, "Timed device " + getDeviceName() + " worker");
        worker.setDaemon(false);
        worker.setPriority(Thread.NORM_PRIORITY);
        worker.start();
    }
    long nanosPerInterval; // How long to wait between pauses
    long cyclesPerInterval; // How many cycles to wait until a pause interval
    long nextSync; // When was the last pause?

    public final void setSpeed(long cyclesPerSecond) {
        cyclesPerInterval = cyclesPerSecond / 60L;
        nanosPerInterval = (long) ((double) cyclesPerInterval * NANOS_PER_SECOND / (double) cyclesPerSecond);
//        System.out.println("Will pause " + nanosPerInterval + " nanos every " + cyclesPerInterval + " cycles");
        cycleTimer = 0;
        resetSyncTimer();
    }
    long skip = 0;
    long wait = 0;

    public final void resetSyncTimer() {
        nextSync = System.nanoTime() + nanosPerInterval;
        cycleTimer = 0;
    }

    public void enableTempMaxSpeed() {
        tempSpeedDuration = TEMP_SPEED_MAX_DURATION;
    }

    public void disableTempMaxSpeed() {
        tempSpeedDuration = 0;
        resetSyncTimer();
    }

    protected void resync() {
        if (++cycleTimer >= cyclesPerInterval) {
            cycleTimer = 0;
            if (maxspeed || tempSpeedDuration > 0) {
                if (tempSpeedDuration > 0) {
                    tempSpeedDuration -= cyclesPerInterval;
                }
                resetSyncTimer();
                return;
            }
            if (clock != null && clock.isAlive()) {
                try {
                    clockLock.lock();
                    clockPulse.await();
                } catch (InterruptedException ex) {
                } finally {
                    clockLock.unlock();
                }
            }
        }
    }

    @Override
    public void reconfigure() {
        if (cyclesPerSecond == 0) {
            cyclesPerSecond = defaultCyclesPerSecond();
        }
        setSpeed(cyclesPerSecond);
    }

    public abstract long defaultCyclesPerSecond();
}
