package jace.hardware;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.locks.LockSupport;

/**
 * Useful reading: http://www.doc.ic.ac.uk/~ih/doc/stepper/others/example3/diskii_specs.html
 * @author brobert
 */
public class DiskIIDrive {

    FloppyDisk disk;
    // Number of milliseconds to wait between last write and update to disk image
    public static long WRITE_UPDATE_DELAY = 1000;
    // Flag to halt if any writes to floopy occur when updating physical disk image
    boolean diskUpdatePending = false;
    // Last time of write operation
    long lastWriteTime;
    // Managed thread to update disk image in background
    Thread writerThread;
    private final byte[][] driveHeadStepDelta = {
        {0, 0, 1, 1, 0, 0, 1, 1, -1, -1, 0, 0, -1, -1, 0, 0}, // phase 0
        {0, -1, 0, -1, 1, 0, 1, 0, 0, -1, 0, -1, 1, 0, 1, 0}, // phase 1
        {0, 0, -1, -1, 0, 0, -1, -1, 1, 1, 0, 0, 1, 1, 0, 0}, // phase 2
        {0, 1, 0, 1, -1, 0, -1, 0, 0, 1, 0, 1, -1, 0, -1, 0}}; // phase 3
    private int halfTrack;
    private int trackStartOffset;
    private int nibbleOffset;
    private boolean writeMode;
    private boolean driveOn;
    private int magnets;
    byte latch;
    private int spinCount;
    Set<Integer> dirtyTracks;

    public void reset() {
        driveOn = false;
        magnets = 0;
        dirtyTracks = new HashSet<Integer>();
        diskUpdatePending = false;
    }

    void step(int register) {
        // switch drive head stepper motor magnets on/off
        int magnet = (register >> 1) & 0x3;
        magnets &= ~(1 << magnet);
        magnets |= ((register & 0x1) << magnet);

        // step the drive head according to stepper magnet changes
        if (driveOn) {
            int delta = driveHeadStepDelta[halfTrack & 0x3][magnets];
            if (delta != 0) {
                int newHalfTrack = halfTrack + delta;
                if (newHalfTrack < 0) {
                    newHalfTrack = 0;
                } else if (newHalfTrack > FloppyDisk.HALF_TRACK_COUNT) {
                    newHalfTrack = FloppyDisk.HALF_TRACK_COUNT;
                }
                if (newHalfTrack != halfTrack) {
                    halfTrack = newHalfTrack;
                    trackStartOffset = (halfTrack >> 1) * FloppyDisk.TRACK_NIBBLE_LENGTH;
                    if (trackStartOffset >= FloppyDisk.DISK_NIBBLE_LENGTH)
                        trackStartOffset = FloppyDisk.DISK_NIBBLE_LENGTH - FloppyDisk.TRACK_NIBBLE_LENGTH;
                    nibbleOffset = 0;

                    //System.out.printf("new half track %d\n", currentHalfTrack);
                }
            }
        }
    }

    void setOn(boolean b) {
        driveOn = b;
    }
    
    boolean isOn() {
        return driveOn;
    }

    byte readLatch() {
        byte result = 0x07f;
        if (!writeMode) {
            spinCount = (spinCount + 1) & 0x0F;
            if (spinCount > 0) {
                if (disk != null) {
                    result = disk.nibbles[trackStartOffset + nibbleOffset++];
                } else {
                    result = (byte) 0x0ff;
                }
            }
            if (nibbleOffset >= FloppyDisk.TRACK_NIBBLE_LENGTH) {
                nibbleOffset = 0;
            }
        } else {
            spinCount = (spinCount + 1) & 0x0F;
            if (spinCount > 0) {
                result = (byte) 0x080;
            }
        }
        return result;
    }

    void write() {
        if (writeMode) {
            while (diskUpdatePending) {
                // If another thread requested writes to block (e.g. because of disk activity), wait for it to finish!
                LockSupport.parkNanos(1000);
            }
            if (disk != null) {
                dirtyTracks.add(trackStartOffset / FloppyDisk.TRACK_NIBBLE_LENGTH);
                disk.nibbles[trackStartOffset + nibbleOffset++] = latch;
                triggerDiskUpdate();
            }

            if (nibbleOffset >= FloppyDisk.TRACK_NIBBLE_LENGTH) {
                nibbleOffset = 0;
        }
    }
    }

    void setLatchValue(byte value) {
        if (writeMode) {
            latch = value;
        } else {
            latch = (byte) 0xFF;
        }
    }

    void setReadMode() {
        writeMode = false;
    }

    void setWriteMode() {
        writeMode = true;
    }

    private void updateDisk() {

        // Signal disk update is underway
        diskUpdatePending = true;
        // Update all tracks as necessary
        for (Integer track : dirtyTracks) {
            if (disk != null) disk.updateTrack(track);
        }
        // Empty out dirty list
        dirtyTracks.clear();
        // Signal disk update is completed
        diskUpdatePending = false;
    }

    private void triggerDiskUpdate() {
        lastWriteTime = System.currentTimeMillis();
        if (writerThread == null || !writerThread.isAlive()) {
            writerThread = new Thread(new Runnable() {
                public void run() {
                    long diff = 0;
                    // Wait until there have been no virtual writes for specified delay time
                    while ((diff = System.currentTimeMillis() - lastWriteTime) < WRITE_UPDATE_DELAY) {
                        // Sleep for difference of time
                        LockSupport.parkNanos(diff * 1000);
                        // Note: In the meantime, there could have been another disk write,
                        // in which case this loop will repeat again as needed.
                    }
                    updateDisk();
                }
            });
            writerThread.start();
        }
    }

    void insertDisk(File diskPath) throws IOException {
        disk = new FloppyDisk(diskPath);
        dirtyTracks = new HashSet<Integer>();
    }

    void ejectDisk() {
        disk = null;
        dirtyTracks = new HashSet<Integer>();
    }
}