package jace.hardware.massStorage;

import jace.Emulator;
import jace.apple2e.MOS65C02;
import jace.config.ConfigurableField;
import jace.config.Name;
import jace.core.Card;
import jace.core.Computer;
import jace.core.Motherboard;
import jace.core.RAMEvent;
import jace.core.RAMEvent.TYPE;
import jace.core.Utility;
import jace.hardware.ProdosDriver;
import jace.hardware.SmartportDriver;
import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.ImageIcon;

/**
 *
 * @author brobert
 */
@Name("Mass Storage Device")
public class CardMassStorage extends Card {
    ImageIcon drive1icon;
    ImageIcon drive2icon;

    public CardMassStorage() {
        drive1icon = Utility.loadIcon("drive-harddisk.png");
        drive2icon = Utility.loadIcon("drive-harddisk.png");
    }

    @Override
    public void setSlot(int slot) {
        super.setSlot(slot);
        drive1icon.setDescription("S"+getSlot()+"D1");
        drive2icon.setDescription("S"+getSlot()+"D2");
    }
    
    @Override
    public String getDeviceName() {
        return "Mass Storage Device";
    }

    // boot0 stores cards*16 of boot device here

    static int SLT16 = 0x02B;
    // "rom" offset where device driver is called by MLI
//    static int DEVICE_DRIVER_OFFSET = 0x042;
    static int DEVICE_DRIVER_OFFSET = 0x0A;
    IDisk disk1;
    IDisk disk2;
    @ConfigurableField(name = "Drive 1", shortName = "d1")
    public File disk1path = null;
    @ConfigurableField(name = "Drive 2", shortName = "d2")
    public File disk2path = null;
    byte[] cardSignature = new byte[]{
        (byte) 0x0a9 /*NOP*/, 0x020, (byte) 0x0a9, 0x00,
        (byte) 0x0a9, 0x03 /*currentDisk cards*/, (byte) 0x0a9, 0x03c /*currentDisk cards*/,
        (byte) 0xd0, 0x07, 0x60, (byte) 0x0b0,
        0x01 /*firmware cards*/, 0x18, (byte) 0x0b0, 0x5a
    };
    Card theCard = this;
    public IDisk currentDisk;
   
    ProdosDriver driver = new ProdosDriver() {
 
        @Override
        public boolean changeUnit(int unit) {
            currentDisk = unit == 0 ? disk1 : disk2;
            return currentDisk != null;
        }

        @Override
        public int getSize() {
            return currentDisk != null ? currentDisk.getSize() : 0;
        }

        @Override
        public boolean isWriteProtected() {
            return currentDisk != null ? currentDisk.isWriteProtected() : true;
        }

        @Override
        public void mliFormat() throws IOException {
            currentDisk.mliFormat();
        }

        @Override
        public void mliRead(int block, int bufferAddress) throws IOException {
            currentDisk.mliRead(block, bufferAddress);
        }

        @Override
        public void mliWrite(int block, int bufferAddress) throws IOException {
            currentDisk.mliWrite(block, bufferAddress);
        }

        @Override
        public Card getOwner() {
            return theCard;
        }
    };

    @Override
    public void reconfigure() {
        try {
            detach();
            if (disk1path == null) {
                if (disk1 != null) {
                    disk1.eject();
                    disk1 = null;
                }
            } else {
                if (disk1 == null) {
                    disk1 = readDisk(disk1path);
                } else if (!disk1.getPhysicalPath().equals(disk1path)) {
                    disk1.eject();
                    disk1 = readDisk(disk1path);
                }
            }
            if (disk2path == null) {
                if (disk2 != null) {
                    disk2.eject();
                    disk2 = null;
                }
            } else {
                if (disk2 == null) {
                    disk2 = readDisk(disk2path);
                } else if (!disk2.getPhysicalPath().equals(disk2path)) {
                    disk2.eject();
                    disk2 = readDisk(disk2path);
                }
            }

            int pc = Computer.getComputer().getCpu().getProgramCounter();
            if (disk1 != null && getSlot() == 7 && (pc == 0x0c65e || pc == 0x0c661)) {
                // If the computer is in a loop trying to boot from cards 6, fast-boot from here instead
                // This is a convenience to boot a hard-drive if the emulator has started waiting for a currentDisk
                disk1.boot0(getSlot());
                Motherboard.cancelSpeedRequest(Motherboard.cards[6]);
            }
            attach();
        } catch (IOException ex) {
            Logger.getLogger(CardMassStorage.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private IDisk readDisk(File f) {
        if (f.isFile()) {
            return new LargeDisk(f);
        } else if (f.isDirectory()) {
            try {
                return new ProdosVirtualDisk(f);
            } catch (IOException ex) {
                System.out.println("Unable to open virtual disk: " + ex.getMessage());
                Logger.getLogger(CardMassStorage.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return null;
    }

    @Override
    public void reset() {
    }

    @Override
    protected void handleC8FirmwareAccess(int register, TYPE type, int value, RAMEvent e) {
        // There is no c8 rom for this card
    }

    @Override
    protected void handleFirmwareAccess(int offset, TYPE type, int value, RAMEvent e) {
        MOS65C02 cpu = (MOS65C02) Computer.getComputer().getCpu();
//                System.out.println(e.getType()+" "+Integer.toHexString(e.getAddress())+" from instruction at  "+Integer.toHexString(cpu.getProgramCounter()));
        if (type.isRead()) {
            if (currentDisk == disk1) {
                Emulator.getFrame().addIndicator(this, drive1icon);
            } else if (currentDisk == disk2) {
                Emulator.getFrame().addIndicator(this, drive2icon);                
            }
            if (disk1 == null && disk2 == null) {
                e.setNewValue(0);
                return;
            }
            if (type == TYPE.EXECUTE) {
                // Virtual functions, handle accordingly
                String error = null;
                if (offset == 0x00) {
                    // NOP unless otherwise specified
                    e.setNewValue(0x0ea);
                    try {
                        if (disk1 != null) {
                            disk1.boot0(getSlot());
                        } else {
                            // Patch for crash on start when no image is mounted
                            e.setNewValue(0x060);
                        }
                        return;
                    } catch (IOException ex) {
                        Logger.getLogger(CardMassStorage.class.getName()).log(Level.SEVERE, null, ex);
                        error = ex.getMessage();
                        // Jump to the basic interpreter for now
                        cpu.setProgramCounter(0x0dfff);
                        int address = 0x0480;
                        for (char c : error.toCharArray()) {
                            Computer.getComputer().getMemory().write(address++, (byte) (c + 0x080), false);
                        }
                    }
                } else {
                    if (offset == DEVICE_DRIVER_OFFSET) {
                        driver.handleMLI();
                    } else if (offset == DEVICE_DRIVER_OFFSET + 3) {
                        smartport.handleSmartport();
                    } else {
                        System.out.println("Call to unknown handler " + Integer.toString(e.getAddress(), 16) + "-- returning");
                    }
                    /* act like RTS was called */
                    e.setNewValue(0x060);
                }
            }
            if (offset < 16) {
                e.setNewValue(cardSignature[offset]);
            } else {
                switch (offset) {
                    // Disk capacity = 65536 blocks
                    case 0x0FC:
                        e.setNewValue(0x0ff);
                        break;
                    case 0x0FD:
                        e.setNewValue(0x07f);
                        break;
                    // Status bits
                    case 0x0FE:
                        e.setNewValue(0x0D7);
                        break;
                    case 0x0FF:
                        e.setNewValue(DEVICE_DRIVER_OFFSET);
                }
            }
        }
    }

    @Override
    protected void handleIOAccess(int register, TYPE type, int value, RAMEvent e) {
        // Ignore IO registers
    }

    @Override
    public void tick() {
        // Nothing is done per CPU cycle
    }

    SmartportDriver smartport = new SmartportDriver(){
        @Override
        public boolean changeUnit(int unitNumber) {
            currentDisk = unitNumber == 1 ? disk1 : disk2;
            return currentDisk != null;
        }

        @Override
        public void read(int blockNum, int buffer) throws IOException {
            currentDisk.mliRead(blockNum, buffer);
        }

        @Override
        public void write(int blockNum, int buffer) throws IOException {
            currentDisk.mliWrite(blockNum, buffer);
        }

        @Override
        public ERROR_CODE returnStatus(int dataBuffer, int[] params) {
            return ERROR_CODE.NO_ERROR;
        }        
    };
}
